(** Time-stamp: <modified the 21/08/2008 (at 15:45) by Erwan Jahier> *)

(**********************************************************
Sous-module pour SyntaxTab 
************************************************************

   Une SymbolTab.t est une association ``contextuelle'' entre
des ident simple (string) et ce  quoi ils correspondent.
   N.B. il est inutile de grer les idents absolus (pack+nom)
qui sont par dfinition NON-CONTEXTUELS.
   Dans une table, on a 3 espaces de noms, un par nature d'item
(const, type et oper).

***********************************************************)
open SyntaxTreeCore
open Lxm

(** Symbol table elements.

Note that the list of static params does not make sens for types
and consts (Well, maybe it could for types?). 
*)
type 'a elt =
  | Local of 'a
  | Imported of Ident.long * static_param srcflagged list

type t

val create : unit -> t
val dump : t -> unit

(* Manip de SymbolTab.t *)

(* Raise a proper compil error message if not found *)
val find_type :  t -> Ident.t -> Lxm.t -> (type_info  Lxm.srcflagged) elt
val find_const : t -> Ident.t -> Lxm.t -> (const_info Lxm.srcflagged) elt
val find_node :  t -> Ident.t -> Lxm.t -> (node_info  Lxm.srcflagged) elt

val find_pack_of_type  : t -> Ident.t -> Lxm.t -> Ident.pack_name
val find_pack_of_const : t -> Ident.t -> Lxm.t -> Ident.pack_name

(* Ajout de nom d'item imports (via uses) *)
val add_import_const : t -> Ident.pack_name -> Ident.t -> Ident.long -> unit
val add_import_type  : t -> Ident.t -> Ident.long -> unit
val add_import_node  : t -> Ident.t -> Ident.long -> 
  static_param srcflagged list -> unit

(** Add local items declaration *)
val add_type  : t -> Ident.pack_name -> Ident.t -> type_info  Lxm.srcflagged -> unit
val add_const : t -> Ident.pack_name -> Ident.t -> const_info Lxm.srcflagged -> unit
val add_node  : t -> Ident.t -> node_info  Lxm.srcflagged -> unit

(* Itrer sur les items *)

val iter_types: t -> (Ident.t -> (type_info Lxm.srcflagged) elt -> unit) -> unit
val iter_consts: t ->(Ident.t -> (const_info Lxm.srcflagged) elt -> unit) -> unit
val iter_nodes : t ->(Ident.t -> (node_info Lxm.srcflagged) elt -> unit) -> unit

